
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <iconv.h>

#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>

#include "sfnt.h"

enum {
	CK_OTHER = 0,
	CK_ASCII,
	CK_ASCII_OR_SJIS2,
	CK_SJIS2,
	CK_SJIS1_OR_SJIS2,
	CK_SJISHK_OR_SJIS2,
} ;

enum {
	CODE_UNKNOWN,
	CODE_ASCII,
	CODE_SJIS,
} ;

int
ckind(unsigned char c)
{
	if (0x20 <= c && c <= 0x3f) return CK_ASCII ;
	if (0x40 <= c && c <= 0x7e) return CK_ASCII_OR_SJIS2 ;
	if (c == 0x80)              return CK_SJIS2 ;
	if (0x81 <= c && c <= 0x9f) return CK_SJIS1_OR_SJIS2 ;
	if (c == 0xa0)              return CK_SJIS2 ;
	if (0xa1 <= c && c <= 0xdf) return CK_SJISHK_OR_SJIS2 ;
	if (0xe0 <= c && c <= 0xfc) return CK_SJIS1_OR_SJIS2 ;
	return CK_OTHER ;
}

int
sjis_check(unsigned char * str, int len)
{
	int i ;
	int state = 0 ;
	
	int tr_table[24] = {
		3, 3, 3, 3,
		0, 3, 2, 3,
		0, 2, 2, 3,
		3, 2, 3, 3,
		1, 2, 1, 3,
		2, 2, 2, 3,
	} ;
	
	for (i=0; i<len; i++) {
		int k = ckind(str[i]) ;
		int nextstate = tr_table[4*k + state] ;
		state = nextstate ;
	}
	
	switch (state) {
	case 0: return CODE_ASCII ;
	case 1: return CODE_UNKNOWN ;
	case 2: return CODE_SJIS ;
	case 3: return CODE_UNKNOWN ;
	}
	return CODE_UNKNOWN ;
}



unsigned char *
convert_UTF16BE_to_ASCII(unsigned char * str, int length)
{
	iconv_t cd ;
	char buf[256] ;
	size_t inleft ;
	size_t outleft ;
	int outlen ;
	char * inptr ;
	char * outptr ;
	int ret ;
	
	cd = iconv_open("ASCII", "UTF-16BE") ;
	if (cd < 0) {
		perror("iconv_open") ;
		return NULL ;
	}
	
	outptr = buf ;
	outleft = 255 ;
	inptr = (char *)str ;
	inleft = length ;
	
	ret = iconv(cd, &inptr, &inleft, &outptr, &outleft) ;
	if (ret < 0) {
		perror("iconv") ;
		iconv_close(cd) ;
		return NULL ;
	}
	
	outlen = 255 - outleft ;
	buf[outlen] = 0 ;
	
	iconv_close(cd) ;
	
	return (unsigned char *)strdup(buf) ;
}



int
fix_dynafont_cmap_table(struct sfnt_cmap_table * cmap)
{
	int i ;
	for (i=0; i<cmap->numberSubtables; i++) {
		struct sfnt_cmap_subtable * stbl = cmap->subtables + i ;
		
		if (stbl->platformID == TT_PLATFORM_MACINTOSH &&
			stbl->platformSpecificID == TT_MAC_ID_ROMAN &&
			stbl->format == 6 &&
			read_uint16(stbl->entity+8) == 256) {
			
			printf("fix.\n") ;
			write_uint16(stbl->entity+8, 127) ;
		}
	}
	return 1 ;
}

int
fix_dynafont_fontfamily(int count, struct sfnt_name_entry * names)
{
	int ret = 0 ;
	struct sfnt_name_entry * ne ;
	
	ne = find_sfnt_name_entry(count, names,
							  TT_PLATFORM_MACINTOSH, TT_MAC_ID_ROMAN,
							  TT_MAC_LANGID_ENGLISH, TT_NAME_ID_FONT_FAMILY) ;
	
	if (ne) {
		if (sjis_check(ne->entity, ne->length) == CODE_SJIS) {
			struct sfnt_name_entry * subs =
				find_sfnt_name_entry(count, names,
									 TT_PLATFORM_MICROSOFT, TT_MS_ID_UNICODE_CS,
									 TT_MS_LANGID_ENGLISH_UNITED_STATES, TT_NAME_ID_FONT_FAMILY) ;
			unsigned char * subs_family ;
			
			if (subs == NULL) {
				fprintf(stderr, "Can't find substitute.\n") ;
				return -1 ;
			}
			
			subs_family = convert_UTF16BE_to_ASCII(subs->entity, subs->length) ;
			if (subs_family == NULL) {
				fprintf(stderr, "Failed to convert substitute to ASCII.\n") ;
				return -1 ;
			}
			printf("%s\n", subs_family) ;
			
			free(ne->entity) ;
			ne->entity = subs_family ;
			ne->length = strlen((char *)subs_family) ;
			
			ret = 1 ;
		}
	}
	
	ne = find_sfnt_name_entry(count, names,
							  TT_PLATFORM_MACINTOSH, TT_MAC_ID_ROMAN,
							  TT_MAC_LANGID_ENGLISH, TT_NAME_ID_FULL_NAME) ;
	
	if (ne) {
		if (sjis_check(ne->entity, ne->length) == CODE_SJIS) {
			struct sfnt_name_entry * subs =
				find_sfnt_name_entry(count, names,
									 TT_PLATFORM_MICROSOFT, TT_MS_ID_UNICODE_CS,
									 TT_MS_LANGID_ENGLISH_UNITED_STATES, TT_NAME_ID_FULL_NAME) ;
			unsigned char * subs_family ;
			
			if (subs == NULL) {
				fprintf(stderr, "Can't find substitute.\n") ;
				return -1 ;
			}
			
			subs_family = convert_UTF16BE_to_ASCII(subs->entity, subs->length) ;
			if (subs_family == NULL) {
				fprintf(stderr, "Failed to convert substitute to ASCII.\n") ;
				return -1 ;
			}
			printf("%s\n", subs_family) ;
			
			free(ne->entity) ;
			ne->entity = subs_family ;
			ne->length = strlen((char *)subs_family) ;
		
			ret = 1 ;
		}
	}
	
	return ret ;
}

int
fix_dynafont_post_table(unsigned char * storage, uint32_t length)
{
	uint32_t isFixedPitch ;
	
	isFixedPitch = read_uint32(storage + 12) ;
	
	if (isFixedPitch) {
		write_uint32(storage + 12, 0) ;
	}
	
	return 1 ;
}


int
main(int argc, char ** argv)
{
	int ret ;
	int fd ;
	uint32_t j ;
	uint32_t length ;
	int need_to_save = 0 ;
	// struct ttc_header ttch ;
	struct stat st ;
	unsigned char * data ;
	struct sfnt_face_info * finfo ;
	
	// uint32_t face_offsets[] = {0} ;
	
	if (argc < 3) exit(1) ;
	
	fd = open(argv[1], O_RDONLY) ;
	if (fd < 0) {
		perror("open") ;
		exit(1) ;
	}
	
	ret = fstat(fd, &st) ;
	if (ret < 0) {
		perror("fstat") ;
		exit(1) ;
	}
	
	length = st.st_size ;
	data = mmap(0, length, PROT_READ, MAP_PRIVATE, fd, 0) ;
	if (data == MAP_FAILED) {
		perror("mmap") ;
		exit(1) ;
	}
	
	close(fd) ;
	
	finfo = parse_sfnt_face_info(data, length) ;
	
	for (j=0; j<finfo->header.numtables; j++) {
		uint32_t tag = read_uint32(finfo->tables[j].tag) ;
		if (tag == 0x636d6170) {
			struct sfnt_cmap_table * cmap ;
			struct uchar_and_length ctbl ;
			printf("parsing cmap table\n") ;
			cmap = parse_sfnt_cmap_table(finfo->tables[j].entity, finfo->tables[j].length) ;
			
			fix_dynafont_cmap_table(cmap) ;
			
			ctbl = render_sfnt_cmap_table(cmap) ;
			finfo->tables[j].length = ctbl.length ;
			finfo->tables[j].padded_length = ((ctbl.length+3)/4)*4 ;
			finfo->tables[j].entity = ctbl.value ;
			
		} else if (tag == 0x6e616d65) {
			int count ;
			int es ;
			struct sfnt_name_entry * names ;
			struct uchar_and_length ntbl ;
			uint32_t offset = finfo->tables[j].offset ;
			uint32_t length = finfo->tables[j].length ;
			struct name_and_count ret ;
			
			ret = parse_sfnt_name_table(data + offset, length) ;
			
			count = ret.count ;
			names = ret.value ;
			
			es = fix_dynafont_fontfamily(count, names) ;
			if (es == 0) continue ;
			
			if (es < 0) {
				printf("Conversion failed.\n") ;
				exit(2) ;
			}
			
			need_to_save = 1 ;
			
			ntbl = render_sfnt_name_table(count, names) ;
			
			finfo->tables[j].length = ntbl.length ;
			finfo->tables[j].padded_length = ((ntbl.length+3)/4)*4 ;
			finfo->tables[j].entity = ntbl.value ;
		} else if (tag == 0x706f7374) {
			unsigned char * data ;
			uint32_t length = finfo->tables[j].length ;
			data = malloc(length) ;
			memcpy(data, finfo->tables[j].entity, length) ;
			finfo->tables[j].entity = data ;
			
			fix_dynafont_post_table(data, length) ;
		}
	}
	
	write_sfnt_face(argv[2], finfo) ;
	
	return 0 ;
}

