
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ttc.h"

void
read_tag(sfnt_tag_t tag, unsigned char * data)
{
	tag[0] = data[0] ;
	tag[1] = data[1] ;
	tag[2] = data[2] ;
	tag[3] = data[3] ;
}

uint32_t
read_uint32(unsigned char * data)
{
	return (data[0]<<24)|(data[1]<<16)|(data[2]<<8)|data[3] ;
}

uint16_t
read_uint16(unsigned char * data)
{
	return (data[0]<<8)|(data[1]) ;
}

void
write_uint32(unsigned char * data, uint32_t x)
{
	data[0] = (x&0x0ff000000)>>24 ;
	data[1] = (x&0x000ff0000)>>16 ;
	data[2] = (x&0x00000ff00)>> 8 ;
	data[3] = (x&0x0000000ff)     ;
}

void
write_uint16(unsigned char * data, uint16_t x)
{
	data[0] = (x&0x0ff00)>>8 ;
	data[1] = (x&0x000ff)    ;
}

uint32_t CalcTableChecksum(unsigned char * data, uint32_t numberOfBytesInTable)
{
	uint32_t i ;
    uint32_t sum = 0;
    uint32_t nLongs = (numberOfBytesInTable + 3) / 4;
	
	for (i=0; i<nLongs*4; i += 4) {
		sum += read_uint32(data+i) ;
	}
	
    return sum;
}

struct ttc_info *
create_ttc_info(uint32_t version)
{
	struct ttc_info * ttcf ;
	
	ttcf = malloc(sizeof(struct ttc_info)) ;
	
	ttcf->version = version ;
	ttcf->face_count = 0 ;
	ttcf->numtables = 0 ;
	
	ttcf->face_capa = 4 ;
	ttcf->faces = malloc(sizeof(struct sfnt_face_info)*4) ;
	
	ttcf->table_capa = 32 ;
	ttcf->tables = malloc(sizeof(struct sfnt_table)*32) ;
	
	return ttcf ;
}

int
ttc_find_table_by_offset(struct ttc_info * ttcf, sfnt_tag_t tag, uint32_t offset)
{
	int i ;
	for (i=0; i<ttcf->numtables; i++) {
		struct sfnt_table * tbl = ttcf->tables + i ;
		
		if (read_uint32(tag) != read_uint32(tbl->tag)) continue ;
		if (offset != tbl->offset) continue ;
		
		return i ;
	}
	return -1 ;
}

int
ttc_find_table(struct ttc_info * ttcf, sfnt_tag_t tag, uint32_t length, unsigned char * entity)
{
	int i ;
	
	for (i=0; i<ttcf->numtables; i++) {
		struct sfnt_table * tbl = ttcf->tables + i ;
		int equal_p = 1 ;
		int k ;
		
		if (read_uint32(tag) != read_uint32(tbl->tag)) continue ;
		if (length != tbl->length) continue ;
		
		for (k=0; k<length; k++) {
			if (entity[k] != tbl->entity[k]) {
				equal_p = 0 ;
				break ;
			}
		}
		
		if (equal_p) return i ;
	}
	
	return -1 ;
}

int
ttc_add_table_direct(struct ttc_info * ttcf, sfnt_tag_t tag, uint32_t offset, uint32_t length, unsigned char * entity)
{
	int idx ;
	struct sfnt_table * tbl ;
	
	idx = ttcf->numtables++ ;
	if (ttcf->numtables > ttcf->table_capa) {
		uint16_t new_capa = ttcf->table_capa + 32 ;
		struct sfnt_table * new_tables = realloc(ttcf->tables, sizeof(struct sfnt_table)*new_capa) ;
		ttcf->table_capa = new_capa ;
		ttcf->tables = new_tables ;
	}
	
	tbl = ttcf->tables + idx ;
	write_uint32(tbl->tag, read_uint32(tag)) ;
	tbl->csum = 0 ;
	tbl->offset = offset ;
	tbl->length = length ;
	tbl->padded_length = (length+3)/4*4 ;
	tbl->entity = malloc(tbl->padded_length) ;
	
	memcpy(tbl->entity, entity, length) ;
	
	return idx ;
}

int
ttc_add_table(struct ttc_info * ttcf, sfnt_tag_t tag, uint32_t length, unsigned char * entity)
{
	int idx = ttc_find_table(ttcf, tag, length, entity) ;
	
	if (idx < 0) return ttc_add_table_direct(ttcf, tag, 0, length, entity) ;
	
	return idx ;
}

struct ttc_info *
ttc_parse_binary(unsigned char * data, uint32_t length)
{
	int i, j ;
	uint32_t rpos ;
	uint32_t version ;
	uint32_t count ;
	struct ttc_info * ttcf ;
	
	if (read_uint32(data+0) != 0x74746366) {
		fprintf(stderr, "Not a TTC data.\n") ;
		return NULL ;
	}
	
	version = read_uint32(data + 4) ;
	count   = read_uint32(data + 8) ;
	rpos = 12 ;
	
	ttcf = malloc(sizeof(struct ttc_info)) ;
	
	ttcf->version = version ;
	ttcf->face_count = count ;
	ttcf->numtables = 0 ;
	
	ttcf->face_capa = count + 4 ;
	ttcf->faces = malloc(sizeof(struct sfnt_face_info)*4) ;
	
	ttcf->table_capa = 32 ;
	ttcf->tables = malloc(sizeof(struct sfnt_table)*32) ;
	
	for (i=0; i<count; i++) {
		uint32_t offset = read_uint32(data+rpos) ;
		rpos += 4 ;
		ttcf->faces[i].offset = offset ;
	}
	
	for (i=0; i<count; i++) {
		struct sfnt_face_info * face = ttcf->faces + i ;
		rpos = face->offset ;
		
		face->header.sctype     = read_uint32(data + rpos +  0) ;
		face->header.numtables  = read_uint16(data + rpos +  4) ;
		face->header.srange     = read_uint16(data + rpos +  6) ;
		face->header.entrysel   = read_uint16(data + rpos +  8) ;
		face->header.rangeshift = read_uint16(data + rpos + 10) ;
		
		rpos += 12 ;
		
		face->table_capa = face->header.numtables ;
		face->table_indexes = malloc(sizeof(int)*face->table_capa) ;
		
		for (j=0; j<face->header.numtables; j++) {
			int idx ;
			uint32_t checksum ;
			uint32_t offset ;
			uint32_t length ;
			sfnt_tag_t tag ;
			
			read_tag(tag, data + rpos + 0) ;
			checksum = read_uint32(data + rpos + 4) ;
			offset   = read_uint32(data + rpos + 8) ;
			length   = read_uint32(data + rpos +12) ;
			rpos += 16 ;
			
			idx = ttc_find_table_by_offset(ttcf, tag, offset) ;
			if (idx < 0) {
				idx = ttc_add_table_direct(ttcf, tag, offset, length, data + offset) ;
				face->table_indexes[j] = idx ;
			} else {
				face->table_indexes[j] = idx ;
			}
		}
	}
	
	return ttcf ;
}

int
ttc_create_face(struct ttc_info * ttcf)
{
	int idx ;
	struct sfnt_face_info * finfo ;
	
	idx = ttcf->face_count++ ;
	if (ttcf->face_count > ttcf->face_capa) {
		uint32_t new_capa = ttcf->face_capa + 8 ;
		struct sfnt_face_info * new_faces = realloc(ttcf->faces, sizeof(struct sfnt_face_info)*new_capa) ;
		ttcf->face_capa = new_capa ;
		ttcf->faces = new_faces ;
	}
	
	finfo = ttcf->faces + idx ;
	
	finfo->offset = 0 ;
	finfo->header.sctype = 0x00010000 ;
	finfo->header.numtables = 0 ;
	finfo->header.srange = 256 ;
	finfo->header.entrysel = 4 ;
	finfo->header.rangeshift = 0 ;
	
	finfo->table_capa = 16 ;
	finfo->table_indexes = calloc(16, sizeof(int)) ;
	
	return idx ;
}

int
ttc_add_table_to_face(struct ttc_info * ttcf, int fidx, sfnt_tag_t tag, uint32_t length, unsigned char * entity)
{
	int idx ;
	struct sfnt_face_info * face ;
	int tidx ;
	
	tidx = ttc_add_table(ttcf, tag, length, entity) ;
	face = ttcf->faces + fidx ;
	
	idx = face->header.numtables++ ;
	if (face->header.numtables > face->table_capa) {
		uint16_t new_capa = face->table_capa + 8 ;
		int * new_table_indexes = realloc(face->table_indexes, sizeof(int)*new_capa) ;
		face->table_capa = new_capa ;
		face->table_indexes = new_table_indexes ;
	}
	
	face->table_indexes[idx] = tidx ;
	
	return tidx ;
}

void
ttc_calculate_checksums(struct ttc_info * ttcf)
{
	int i ;
	for (i=0; i<ttcf->numtables; i++) {
		ttcf->tables[i].csum = CalcTableChecksum(ttcf->tables[i].entity, ttcf->tables[i].length) ;
	}
}

int
ttc_calculate_length_and_offsets(struct ttc_info * ttcf, int headersize)
{
	int i ;
	int length = headersize ;
	
	for (i=0; i<ttcf->numtables; i++) {
		ttcf->tables[i].offset = length ;
		length += ttcf->tables[i].padded_length ;
	}
	
	return length ;
}


int
ttc_write_to_file(struct ttc_info * ttcf, FILE * fp)
{
	size_t ret ;
	int i, j ;
	int wlen = 0 ;
	int wpos ;
	int headersize ;
	unsigned char * header ;
	
	ttc_calculate_checksums(ttcf) ;
	
	headersize = 12 + ttcf->face_count * 4 ;
	
	for (i=0; i<ttcf->face_count; i++) {
		ttcf->faces[i].offset = headersize ;
		headersize += 12 + ttcf->faces[i].header.numtables * 16 ;
	}
	
	ttc_calculate_length_and_offsets(ttcf, headersize) ;
	
	header = malloc(headersize) ;
	
	write_uint32(header + 0, 0x74746366) ;
	write_uint32(header + 4, ttcf->version) ;
	write_uint32(header + 8, ttcf->face_count) ;
	
	wpos = 12 ;
	for (i=0; i<ttcf->face_count; i++) {
		write_uint32(header + wpos, ttcf->faces[i].offset) ;
		wpos += 4 ;
	}
	
	for (i=0; i<ttcf->face_count; i++) {
		struct sfnt_face_info * face = ttcf->faces + i ;
		write_uint32(header + wpos +  0, face->header.sctype) ;
		write_uint16(header + wpos +  4, face->header.numtables) ;
		write_uint16(header + wpos +  6, face->header.srange) ;
		write_uint16(header + wpos +  8, face->header.entrysel) ;
		write_uint16(header + wpos + 10, face->header.rangeshift) ;
		wpos += 12 ;
		
		for (j=0; j<face->header.numtables; j++) {
			struct sfnt_table * tbl = ttcf->tables + face->table_indexes[j] ;
			write_uint32(header + wpos +  0, read_uint32(tbl->tag)) ;
			write_uint32(header + wpos +  4, tbl->csum) ;
			write_uint32(header + wpos +  8, tbl->offset) ;
			write_uint32(header + wpos + 12, tbl->length) ;
			wpos += 16 ;
		}
	}
	
	ret = fwrite(header, 1, headersize, fp) ;
	if (ret < headersize) {
		perror("fwrite") ;
		return -1 ;
	}
	wlen += ret ;
	
	for (i=0; i<ttcf->numtables; i++) {
		uint32_t length = ttcf->tables[i].padded_length ;
		ret = fwrite(ttcf->tables[i].entity, 1, length, fp) ;
		if (ret < length) {
			perror("fwrite") ;
			return -1 ;
		}
		wlen += ret ;
	}
	
	return wlen ;
}

int
ttc_write_face_to_file(struct ttc_info * ttcf, int fidx, FILE * fp)
{
	size_t ret ;
	int wlen = 0 ;
	int i ;
	uint32_t length = 0 ;
	struct sfnt_face_info * face ;
	int headersize ;
	unsigned char * header ;
	
	ttc_calculate_checksums(ttcf) ;
	
	face = ttcf->faces + fidx ;
	headersize = 12 + face->header.numtables * 16 ;
	
	length = headersize ;
	for (i=0; i<face->header.numtables; i++) {
		struct sfnt_table * tbl = ttcf->tables + face->table_indexes[i] ;
		tbl->offset = length ;
		length += tbl->padded_length ;
	}
	
	header = malloc(headersize) ;
	write_uint32(header +  0, face->header.sctype) ;
	write_uint16(header +  4, face->header.numtables) ;
	write_uint16(header +  6, face->header.srange) ;
	write_uint16(header +  8, face->header.entrysel) ;
	write_uint16(header + 10, face->header.rangeshift) ;
	
	for (i=0; i<face->header.numtables; i++) {
		struct sfnt_table * tbl = ttcf->tables + face->table_indexes[i] ;
		write_uint32(header + 12 + i*16 +  0, read_uint32(tbl->tag)) ;
		write_uint32(header + 12 + i*16 +  4, tbl->csum) ;
		write_uint32(header + 12 + i*16 +  8, tbl->offset) ;
		write_uint32(header + 12 + i*16 + 12, tbl->length) ;
	}
	
	ret = fwrite(header, 1, headersize, fp) ;
	if (ret < headersize) {
		perror("fwrite") ;
		return -1 ;
	}
	wlen += ret ;
	
	for (i=0; i<face->header.numtables; i++) {
		struct sfnt_table * tbl = ttcf->tables + face->table_indexes[i] ;
		uint32_t length = tbl->padded_length ;
		ret = fwrite(tbl->entity, 1, length, fp) ;
		if (ret < length) {
			perror("fwrite") ;
			return -1 ;
		}
		wlen += ret ;
	}
	
	return ret ;
}

